/******************************************************************************/
/* tlr                                                                        */
/* (c) 2006 jamie twycross, jpt AT cs.nott.ac.uk                              */
/* released under gnu gpl v2                                                  */
/******************************************************************************/

#include <tissue/tissue.h>
#include <tissue/syscall.h>

#include <stdlib.h>
#include <unistd.h>
#include <err.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <asm/unistd.h>
#include <math.h>

/* model parameters */
unsigned int max_antigen;
unsigned int antigen_multiplier;
unsigned int max_cytokines;
unsigned int max_cells;
unsigned int cell_update_rate;
unsigned int num_cells_1;
unsigned int cell_lifespan_1;
unsigned int num_antigen_1;
unsigned int num_antigen_receptors_1;
unsigned int num_antigen_producers_1;
unsigned int antigen_producer_action_time;
unsigned int num_cytokine_receptors_1;
unsigned int num_cells_2;
unsigned int cell_lifespan_2;
unsigned int num_cell_receptors_2;
unsigned int num_vr_receptors_2;
unsigned int num_progeny_2;
unsigned int cell_lifespan_3;
unsigned int cell_lifespan_4;
unsigned int cell_lifespan_5;
unsigned int ts_port = TS_DEFAULT_PORT;
unsigned int log_tissue = 0;
unsigned int probe_rate = 0;

char parameter_file[1024] = "";
char known_antigen_file[1024] = "";
char known_signals_file[1024] = "";

unsigned int num_known_antigen, num_known_signals;
Antigen **known_antigen;
Signal **known_signals;

unsigned int uniqueid = 1;

/* extract syscall number from antigen */
int extract_sc(Antigen *ag)
{
	static char val[5], *endptr[1];

	memcpy(val, ag->val, 4);  /* four bytes */
	val[4] = '\0';
	return strtol(val, endptr, 16);
}

/* probe compartment periodically + save data to log */
void probe_tissue(Tissue *tissue)
{
	register unsigned int i, j, count;
	static char message[32768], *ptr;

	ptr = message;
	ptr += sprintf(ptr, "tissue");
	for(i = 0; i < tissue->num_cytokines; i++)
		ptr += sprintf(ptr, " %f", tissue->cytokine[i]->val);
	logtime_tissue(message);
	
	for(i = 0; i < tissue->num_cells; i++)
		if(tissue->cell[i]) {
			if(tissue->cell[i]->type == 1) {  /* iDC */
				ptr = message;
				ptr += sprintf(ptr, "cell 1 %d %d", \
						tissue->cell[i]->id, \
						tissue->cell[i]->iterations);
				for(j = 0; j < num_cytokine_receptors_1; j++)
					ptr += sprintf(ptr, " %f", \
							tissue->cell[i]->cytokine_receptor[j]->val);
				for(j = 0; j < tissue->cell[i]->num_antigen; j++)
					if(tissue->cell[i]->antigen[j])
						ptr += sprintf(ptr, " %s", \
								antigentostr(tissue->cell[i]->antigen[j]));
				logtime_tissue(message);
				continue;
			}
			if(tissue->cell[i]->type == 2) {  /* nTC */
				ptr = message;
				ptr += sprintf(ptr, "cell 2 %d %d %s", \
						tissue->cell[i]->id, \
						tissue->cell[i]->iterations, \
						antigentostr(tissue->cell[i]->vr_receptor[0]->lock));
				count = 0;
				for(j = 0; j < num_vr_receptors_2; j++)
					if(tissue->cell[i]->vr_receptor[j]->activated)
						count++;
				ptr += sprintf(ptr, " %d", count);
				count = 0;
				for(j = 0; j < tissue->cell[i]->num_cell_receptors; j++)
					if(tissue->cell[i]->cell_receptor[j]->cell)
						count++;
				ptr += sprintf(ptr, " %d", count);
				logtime_tissue(message);
				continue;
			}
			if(tissue->cell[i]->type == 3) {  /* smDC */
				ptr = message;
				ptr += sprintf(ptr, "cell 3 %d %d", \
						tissue->cell[i]->id, \
						tissue->cell[i]->iterations);
				for(j = 0; j < num_cytokine_receptors_1; j++)
					ptr += sprintf(ptr, " %f", \
							tissue->cell[i]->cytokine_receptor[j]->val);
				for(j = 0; j < tissue->cell[i]->num_antigen; j++)
					if(tissue->cell[i]->antigen[j])
						ptr += sprintf(ptr, " %s", \
								antigentostr(tissue->cell[i]->antigen[j]));
				logtime_tissue(message);
				continue;
			}
			if(tissue->cell[i]->type == 4) {  /* mDC */
				ptr = message;
				ptr += sprintf(ptr, "cell 4 %d %d", \
						tissue->cell[i]->id, \
						tissue->cell[i]->iterations);
				for(j = 0; j < num_cytokine_receptors_1; j++)
					ptr += sprintf(ptr, " %f", \
							tissue->cell[i]->cytokine_receptor[j]->val);
				for(j = 0; j < tissue->cell[i]->num_antigen; j++)
					if(tissue->cell[i]->antigen[j])
						ptr += sprintf(ptr, " %s", \
								antigentostr(tissue->cell[i]->antigen[j]));
				logtime_tissue(message);
				continue;
			}
			if(tissue->cell[i]->type == 5) {  /* aTC */
				sprintf(message, "cell 5 %d %d %s %f", \
						tissue->cell[i]->id, \
						tissue->cell[i]->iterations, \
						antigentostr(tissue->cell[i]->vr_receptor[0]->lock), \
						tissue->cell[i]->cytokine[0]->val);
				logtime_tissue(message);
				continue;
			}
		}
}

/* vr receptor matching callback */
inline double antigen_match(Antigen *lock, Antigen *key)
{
	if(!key || !lock)
		return 0.0;
	/* exact match */
	if(memcmp(lock->val, key->val, lock->len) == 0)
		return 1.0;
	return 0.0;
}

/* create antigen receptor */
/* 4 characters - syscall number as 4-digit hex */
Antigen *random_lock(void)
{
	Antigen *lock = NULL;
	static char line[64];
	static int sc;
	register unsigned int i;

	while(!lock) {
		/* set lock to random syscall number */
		sc = (int) (((double) nsyscalls * rand()) / (RAND_MAX + 1.0));
		snprintf(line, sizeof(line), "%.4x", sc);
		lock = set_antigen(lock, line);
		/* check to see not known antigen */
		for(i = 0; i < num_known_antigen; i++)
			if(antigen_match(lock, known_antigen[i])) {
				lock = NULL;
				break;
			}
	}

	return lock;
}

/* produce a response */
char *response_message(Cell *cell, Antigen *ag)
{
	static char line[256], name[256];
	char *message;

	snprintf(line, sizeof(line), "%d %s(%d) %d %s deny", \
			cell->id, \
			syscall_name(extract_sc(cell->vr_receptor[0]->lock), name), \
			extract_sc(cell->vr_receptor[0]->lock), \
			ag->id, \
			antigentostr(cell->vr_receptor[0]->lock));
	
	/* set message */
	if(!(message = calloc(strlen(line) + 1, sizeof(char))))
		err(1, "couldn't allocate memory for message");
	strncpy(message, line, strlen(line) + 1);

	return message;
}

/* check if cytokine receptor is activated */
int cytokine_receptor_activated(Cytokine_Receptor *ckr)
{
	register unsigned int i;

	if(!ckr || ckr->val == 0.0)
		return 0;  /* unactivated if value 0.0 */
	for(i = 0; i < num_known_signals; i++)
		if((ckr->id == known_signals[i]->id) && \
				(ckr->val == known_signals[i]->val))
			return 0;  /* known value, so unactivated */
	return 1;  /* unknown value, so activated */
}

/* smDC cell cycle */
void cell_cycle_3(Cell *cell)
{
	/* just produce antigen until lifespan reached */
}

/* mDC cell cycle */
void cell_cycle_4(Cell *cell)
{
	/* just produce antigen until lifespan reached */
}

/* aTC cell cycle */
void cell_cycle_5(Cell *cell)
{
	/* just match tissue antigen until lifespan reached */
}

/* iDC cell cycle */
void cell_cycle_1(Cell *cell)
{
	unsigned int i, j;
	
	/* check lifespan + mature into smDC at end of life */
	if(cell->iterations >= cell_lifespan_1) {
		/* only mature if collected antigen */
		for(j = 0; j < cell->num_antigen; j++)
			if(cell->antigen[j])
				break;
		if(j == cell->num_antigen) {
			/* no antigen collected so kill cell (apoptosis) */
			/* really just reset id + iterations */
			cell->id = uniqueid++;  /* set cell ID */
			cell->iterations = 0;  /* reset iterations */
		return;
		}
		/* mature to smDC */
		cell->type = 3;  /* set type to smDC */
		cell->iterations = 0;  /* reset iterations */
		/* activate antigen producers */
		cell->num_antigen_producers = num_antigen_producers_1;
		/* deactivate antigen receptors */
		cell->num_antigen_receptors = 0;
		/* deactivate cytokine receptors */
		cell->num_cytokine_receptors = 0;
		/* set internal cytokine level to 1.0 (semimature) */
		cell->cytokine[0]->val = 1.0;
		/* set cell cycle */
		cell->cycle = cell_cycle_3;
		return;
	}
	
	/* monitor signal level and mature if activated */
	for(i = 0; i < cell->num_cytokine_receptors; i++)
		if(cytokine_receptor_activated(cell->cytokine_receptor[i]))
		{
			/* only mature if collected antigen */
			for(j = 0; j < cell->num_antigen; j++)
				if(cell->antigen[j])
					break;
			if(j == cell->num_antigen)
				return;  /* no antigen collected */
			/* mature to mDC */
			cell->type = 4;  /* set type to mDC */
			cell->iterations = 0;  /* reset iterations */
			/* activate antigen producers */
			cell->num_antigen_producers = num_antigen_producers_1;
			/* deactivate antigen receptors */
			cell->num_antigen_receptors = 0;
			/* deactivate cytokine receptors */
			cell->num_cytokine_receptors = 0;
			/* set internal cytokine level to 2.0 (mature) */
			cell->cytokine[0]->val = 2.0;
			/* set cell cycle */
			cell->cycle = cell_cycle_4;
			return;
		}
}

/* nTC cell cycle */
void cell_cycle_2(Cell *cell)
{
	unsigned int i;
	Antigen *lock;
	
	/* check lifespan */
	if(cell->iterations >= cell_lifespan_2) {
		/* reached end of life so die */
		/* really just create new nTC by randomising vrr */
		lock = random_lock();
		for(i = 0; i < num_vr_receptors_2; i++) {
			free_antigen(cell->vr_receptor[i]->lock);
			cell->vr_receptor[i]->lock = dup_antigen(lock);
		}
		cell->id = uniqueid++;  /* set cell ID */
		cell->iterations = 0;  /* reset iterations */
		return;
	}

	/* check if any vr receptors are activated */
	for(i = 0; i < cell->num_vr_receptors; i++)	{
		if(cell->vr_receptor[i]->activated) {
			/* vr receptors matched */
			if(cell->vr_receptor[i]->cell->type == 3) {
				/* matched cell is smDC */
				/* keep smDC alive */
				cell->vr_receptor[i]->cell->iterations = 1;
				/* die (peripheral tolerance) */
				/* really just create new nTC by randomising vrr */
				lock = random_lock();
				for(i = 0; i < num_vr_receptors_2; i++) {
					free_antigen(cell->vr_receptor[i]->lock);
					cell->vr_receptor[i]->lock = dup_antigen(lock);
  				}
				cell->id = uniqueid++;  /* set cell ID */
				cell->iterations = 0;  /* reset iterations */
				return;
			}
			if(cell->vr_receptor[i]->cell->type == 4) {
				/* matched cell is mDC */
				/* keep mDC alive */
				cell->vr_receptor[i]->cell->iterations = 1;
				/* mature to aTC */
				cell->type = 5;  /* set type to aTC */
				cell->iterations = 0;  /* reset iterations */
				/* set cell cycle */
				cell->cycle = cell_cycle_5;
				/* disable cell receptors */
				cell->num_cell_receptors = 0;
				/* disable vr receptors */
				cell->num_vr_receptors = 0;
				return;				
			}
		}
	}
}

/* create a generic DC */
Cell *create_dc(void)
{
	Cell_Param cell_param;
	Cell *cell;

	/* set parameters for cell creation */
	/* generic dc */
	bzero(&cell_param, sizeof(Cell_Param));
	cell_param.type = 1;  /* type 1 - iDC */
	cell_param.num_antigen = num_antigen_1; /* num of antigen cell stored */
	cell_param.num_cytokines = 1; /* 1 cytokine to show if matured */
	cell_param.num_cytokine_receptors = num_cytokine_receptors_1;/* signal */
	cell_param.num_antigen_receptors = num_antigen_receptors_1;
	cell_param.num_antigen_producers = num_antigen_producers_1;
	cell_param.cycle = cell_cycle_1;

	/* initialise the cell */
	cell = init_cell(&cell_param);
	
	return cell;
}

/* create a generic TC */
Cell *create_tc(void)
{
	Cell_Param cell_param;
	Cell *cell;

	/* set parameters for cell creation */
	/* generic TC */
	bzero(&cell_param, sizeof(Cell_Param));
	cell_param.type = 2;  /* type 2 */
	cell_param.num_cytokines = 1;  /* keep number of matches */
	cell_param.num_cell_receptors = num_cell_receptors_2 * 2;
	cell_param.num_vr_receptors = num_vr_receptors_2;
	cell_param.cycle = cell_cycle_2;

	/* initialise the cell */
	cell = init_cell(&cell_param);
	
	return cell;
}

/* create iDC */
Cell *create_idc(void)
{
	Cell *cell;
	unsigned int i;
	
	/* initialise the cell */
	cell = create_dc();

	/* set cytokine receptor to respond to tissue signal */
	for(i = 0; i < num_cytokine_receptors_1; i++) {
		cell->cytokine_receptor[i]->id = i;
		/* no noise for cytokine receptor */
		cell->cytokine_receptor[i]->noise = 0.0;
	}
	/* set parameters for antign producers */
	for(i = 0; i < num_antigen_producers_1; i++) {
		/* action time */
		cell->antigen_producer[i]->action_time = antigen_producer_action_time;
		/* destroy flag - don't remove antigen from cell store */
		cell->antigen_producer[i]->destroy = 0;
	}
	/* make antigen producers initially inactive */
	cell->num_antigen_producers = 0;

	return cell;
}

/* create nTC */
Cell *create_ntc(void)
{
	Cell *cell;
	unsigned int i;
	Antigen *lock;
	
	/* initialise the cell */
	cell = create_tc();

	/* set cell receptor to respond to smDCs */
	for(i = 0; i < num_cell_receptors_2; i++)
		cell->cell_receptor[i]->type = 3;
	/* set cell receptor to respond to mDCs */
	for(i = 0; i < num_cell_receptors_2; i++)
		cell->cell_receptor[i + num_cell_receptors_2]->type = 4;
	/* set random values for locks of vr receptors */
	lock = random_lock();  /* same lock for all receptors */
	for(i = 0; i < num_vr_receptors_2; i++) {
		cell->vr_receptor[i]->lock = dup_antigen(lock);
		cell->vr_receptor[i]->threshold = 1.0;
		cell->vr_receptor[i]->match = &antigen_match;
	}

	return cell;
}

/* create aTC */
Cell *create_atc(void)
{
	Cell *cell;
	
	/* initialise the cell */
	cell = create_tc();

	cell->type = 5;  /* set type to aTC */
	/* set cell cycle */
	cell->cycle = cell_cycle_5;
	/* disable cell receptors */
	cell->num_cell_receptors = 0;
	/* disable vr receptors */
	cell->num_vr_receptors = 0;
	
	return cell;
}

/* find empty slot in tissue for cell */
unsigned int empty_cell_slot(Tissue *tissue)
{
	register unsigned int ind;

	for(ind = 0; ind < tissue->num_cells; ind++)
		if(!tissue->cell[ind])
			return ind;
	/* fail if tissue full */
	logfatal_tissue("error: no cell slots left\n");
	return ind;
}

/* initialise some cells */
void create_tissue_cells(Tissue *tissue)
{
	unsigned int i, ind;

	i = num_cells_1;
	while(i--) {
		/* find spare slot in tissue to put cell */
		ind = empty_cell_slot(tissue);
		/* initialise the cell */
		tissue->cell[ind] = create_idc();
		/* set cell ID */
		tissue->cell[ind]->id = uniqueid++;  
		/* set cell iterations to random value to avoid lockstep */
		tissue->cell[ind]->iterations = (int) (((double) \
				cell_lifespan_1 * rand()) / (RAND_MAX + 1.0));
	}

	i = num_cells_2;
	while(i--) {
		/* find spare slot in tissue to put cell */
		ind = empty_cell_slot(tissue);
		/* initialise the cell */
		tissue->cell[ind] = create_ntc();
		/* set cell ID */
		tissue->cell[ind]->id = uniqueid++;  
		/* set cell iterations to random value to avoid lockstep */
		tissue->cell[ind]->iterations = (int) (((double) \
				cell_lifespan_2 * rand()) / (RAND_MAX + 1.0));
	}
}

/* read parameters from file */
void read_parameters(char *filename)
{
	read_parameter_uint(filename, "max_antigen", &max_antigen);
	read_parameter_uint(filename, "max_cytokines", &max_cytokines);
	read_parameter_uint(filename, "max_cells", &max_cells);
	read_parameter_uint(filename, "cell_update_rate", &cell_update_rate);
	read_parameter_uint(filename, "num_cells_1", &num_cells_1);
	read_parameter_uint(filename, "cell_lifespan_1", &cell_lifespan_1);
	read_parameter_uint(filename, "num_antigen_1", &num_antigen_1);
	read_parameter_uint(filename, "num_antigen_receptors_1", \
			&num_antigen_receptors_1);
	read_parameter_uint(filename, "num_antigen_producers_1", \
			&num_antigen_producers_1);
	read_parameter_uint(filename, "antigen_producer_action_time", \
			&antigen_producer_action_time);
	read_parameter_uint(filename, "num_cytokine_receptors_1", \
			&num_cytokine_receptors_1);
	read_parameter_uint(filename, "num_cells_2", &num_cells_2);
	read_parameter_uint(filename, "cell_lifespan_2", &cell_lifespan_2);
	read_parameter_uint(filename, "num_cell_receptors_2", \
			&num_cell_receptors_2);
	read_parameter_uint(filename, "num_vr_receptors_2", &num_vr_receptors_2);
	read_parameter_uint(filename, "num_progeny_2", &num_progeny_2);
	read_parameter_uint(filename, "cell_lifespan_3", &cell_lifespan_3);
	read_parameter_uint(filename, "cell_lifespan_4", &cell_lifespan_4);
	read_parameter_uint(filename, "cell_lifespan_5", &cell_lifespan_5);
	read_parameter_uint(filename, "ts_port", &ts_port);
	read_parameter_uint(filename, "log_tissue", &log_tissue);
	read_parameter_uint(filename, "probe_rate", &probe_rate);
	read_parameter_uint(filename, "antigen_multiplier", &antigen_multiplier);
}

/* read known antigen from file */
void read_known_antigen(char *filename)
{
	FILE *fp;
	char line[1024], message[32768], *ptr2;;
	unsigned int i;

	if(!(fp = fopen(filename, "r"))) {
		fprintf(stderr, "error: couldn't open known antigen file\n");
		exit(1);
	}
	/* file format is one antigen per line */
	/* first count total antigen */
	num_known_antigen = 0;
	while(freadline(line, 1024, fp))
		num_known_antigen++;
	/* allocate memory */
	if(!(known_antigen = calloc(num_known_antigen, sizeof(Antigen *)))) {
		fprintf(stderr, "error: couldn't allocate memory for known antigen\n");
		exit(1);
	}
	/* read in antigen */
	rewind(fp);
	ptr2 = message;
	ptr2 += sprintf(ptr2, "using %d known antigen from file %s: ", \
			num_known_signals, filename);
	for(i = 0; i < num_known_antigen; i++) {
		freadline(line, 1024, fp);
		ptr2 += sprintf(ptr2, " %s", line);
		known_antigen[i] = init_antigen(strlen(line));
		set_antigen_fast(known_antigen[i], line);
	}
	/* log number of known antigen */
	logtime_tissue(message);
	
	fclose(fp);
}

/* read known signals from file */
void read_known_signals(char *filename)
{
	FILE *fp;
	char line[1024], *ptr, message[32768], *ptr2;
	unsigned int i;

	if(!(fp = fopen(filename, "r"))) {
		fprintf(stderr, "error: couldn't open known signals file\n");
		exit(1);
	}
	/* file format is two column (signal id, signal value) */
	/* first count total signals */
	num_known_signals = 0;
	while(freadline(line, 1024, fp))
		num_known_signals++;
	/* allocate memory */
	if(!(known_signals = calloc(num_known_signals, sizeof(Signal *)))) {
		fprintf(stderr, "error: couldn't allocate memory for known signals\n");
		exit(1);
	}
	/* read in signals */
	rewind(fp);
	ptr2 = message;
	ptr2 += sprintf(ptr2, "using %d known signals from file %s: ", \
			num_known_signals, filename);
	for(i = 0; i < num_known_signals; i++) {
		freadline(line, 1024, fp);
		ptr2 += sprintf(ptr2, " %s", line);
		/* extract signal id and value */
		ptr = strchr(line, ' ');
		*ptr = '\0';
		known_signals[i] = init_signal();
		known_signals[i]->id = atoi(line);
		known_signals[i]->val = atof(ptr + 1);
	}

	/* log number of known signals */
	logtime_tissue(message);
	fclose(fp);
}

static void parse_command_line(int argc, char **argv)
{
	if(argc == 4) {
		strncpy(parameter_file, argv[1], sizeof(parameter_file));
		strncpy(known_antigen_file, argv[2], sizeof(known_antigen_file));
		strncpy(known_signals_file, argv[3], sizeof(known_signals_file));
		read_parameters(parameter_file);
	} else {
		fprintf(stderr, "error: tlr-lt PARAMS ANTIGEN SIGNALS\n");
		exit(1);
	}
}

int main(int argc, char **argv)
{
	Tissue *tissue;
	register unsigned int i, j, k, ind;
	Antigen *lock;

	parse_command_line(argc, argv);

	/* create a tissue compartment */
	tissue = init_tissue((int16_t) ts_port, max_antigen, max_cytokines, \
			max_cells);

	/* save user specified parameters to log */
	log_parameter_file(parameter_file);

	/* read known training antigen and signals */
	read_known_antigen(known_antigen_file);
	read_known_signals(known_signals_file);

	/* log input data from tissue clients to compartment */
	tissue->log = (short int) log_tissue;

	/* set multiplier for incoming antigen */
	tissue->antigen_multiplier = antigen_multiplier;

	/* initialise some cells in the compartment */
	create_tissue_cells(tissue);

	/* initialise a probe to monitor compartment */
	init_probe(tissue, probe_tissue, probe_rate);

	/* run tlr algorithm */
	while(1) {
		/* let cells interact in tissue */
		step_tissue(tissue);

		/* keep iDC and nTC numbers constant, match aTCs, replace old cells */
        pthread_mutex_lock(tissue->lock);
		for(i = 0; i < max_cells; i++) {
			if(tissue->cell[i]) {
				if(tissue->cell[i]->type == 3) {  /* smDC */
					if(tissue->cell[i]->iterations >= cell_lifespan_3) {
						/* reached end of life, so kill (apoptosis) */
						free_cell(tissue->cell[i]);
						tissue->cell[i] = NULL;
						continue;
					}
				}
				if(tissue->cell[i]->type == 4) {  /* mDC */
					if(tissue->cell[i]->iterations >= cell_lifespan_4) {
						/* reached end of life, so kill (apoptosis) */
						free_cell(tissue->cell[i]);
						tissue->cell[i] = NULL;
						continue;
					}
				}
				if(tissue->cell[i]->type == 5) {  /* aTC */
					if(tissue->cell[i]->iterations >= cell_lifespan_5) {
						/* reached end of life, so kill (apoptosis) */
						free_cell(tissue->cell[i]);
						tissue->cell[i] = NULL;
						continue;
					}
				}
				if((tissue->cell[i]->type == 3) || \
						(tissue->cell[i]->type == 4)) {  /* smDC or mDC */
					if(!tissue->cell[i]->iterations) {
						/* new smDC, so add an iDC to replace */
						/* find spare slot in tissue to put cell */
						ind = empty_cell_slot(tissue);
						/* initialise the cell */
						tissue->cell[ind] = create_idc();
						/* set cell ID */
						tissue->cell[ind]->id = uniqueid++;  
						continue;
					}
				}
				if(tissue->cell[i]->type == 5) {  /* aTC */
					if(!tissue->cell[i]->iterations) {
						/* new aTC, so add an nTC to replace */
						/* find spare slot in tissue to put cell */
						ind = empty_cell_slot(tissue);
						/* initialise the cell */
						tissue->cell[ind] = create_ntc();
						/* set cell ID */
						tissue->cell[ind]->id = uniqueid++;  
						
						/* and proliferate children */
						for(j = 0; j < num_progeny_2; j++) {
							/* find spare slot in tissue to put cell */
							ind = empty_cell_slot(tissue);
							/* initialise the cell */
							tissue->cell[ind] = create_atc();
							/* copy lock */
							lock = tissue->cell[i]->vr_receptor[0]->lock;
							for(k = 0; k < num_vr_receptors_2; k++) {
								tissue->cell[ind]->vr_receptor[k]->lock = \
										dup_antigen(lock);
								tissue->cell[ind]->vr_receptor[k]->threshold = \
										1.0;
								tissue->cell[ind]->vr_receptor[k]->match = \
										&antigen_match;
							}
							/* set cell ID */
							tissue->cell[ind]->id = uniqueid++;
							/* set iterations to 1 */
							tissue->cell[ind]->iterations = 1;
						}
					}
					/* check for antigen matches */
					for(j = 0; j < tissue->num_antigen; j++) {
						if(antigen_match(tissue->cell[i]->vr_receptor[0]->lock, \
								tissue->antigen[j])) {
									logtime_tissue(response_message( \
											tissue->cell[i], \
											tissue->antigen[j]));
									/* keep aTC alive */
									tissue->cell[i]->iterations = 1;
									/* record number of matches */
									tissue->cell[i]->cytokine[0]->val += 1.0;
						}
					}
				}
			}
		}
        pthread_mutex_unlock(tissue->lock);
	
		/* wait a bit */
		usleep(cell_update_rate);
	}

	return EXIT_SUCCESS;
}
